@@ -139,6 +139,7 @@ locate_pointer_expose (GtkWidget *widget,
       cr = gdk_cairo_create (mask);
       locate_pointer_paint (data, cr, FALSE);
       gdk_window_shape_combine_mask (data->window, mask, 0, 0);
+      g_object_unref (mask);
     }
 
   cairo_destroy (cr);
@@ -152,6 +153,8 @@ timeline_frame_cb (GsdTimeline *timeline,
 		   gpointer     user_data)
 {
   GsdLocatePointerData *data = (GsdLocatePointerData *) user_data;
+  GdkScreen *screen;
+  gint cursor_x, cursor_y;
 
   if (gtk_widget_is_composited (data->widget))
     {
@@ -164,6 +167,47 @@ timeline_frame_cb (GsdTimeline *timeline,
       gdk_window_invalidate_rect (data->window, NULL, FALSE);
       data->progress += CIRCLES_PROGRESS_INTERVAL;
     }
+
+  screen = gdk_drawable_get_screen (data->window);
+  gdk_window_get_pointer (gdk_screen_get_root_window (screen),
+			  &cursor_x, &cursor_y, NULL);
+  gdk_window_move (data->window,
+                   cursor_x - WINDOW_SIZE / 2,
+                   cursor_y - WINDOW_SIZE / 2);
+}
+
+static void
+set_transparent_shape (GdkWindow *window)
+{
+  GdkBitmap *mask;
+  cairo_t *cr;
+
+  mask = gdk_pixmap_new (data->window, WINDOW_SIZE, WINDOW_SIZE, 1);
+  cr = gdk_cairo_create (mask);
+
+  cairo_set_source_rgba (cr, 1., 1., 1., 0.);
+  cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
+  cairo_paint (cr);
+
+  gdk_window_shape_combine_mask (data->window, mask, 0, 0);
+  g_object_unref (mask);
+  cairo_destroy (cr);
+}
+
+static void
+unset_transparent_shape (GdkWindow *window)
+{
+  gdk_window_shape_combine_mask (data->window, NULL, 0, 0);
+}
+
+static void
+composited_changed (GtkWidget            *widget,
+                    GsdLocatePointerData *data)
+{
+  if (!gtk_widget_is_composited (widget))
+    set_transparent_shape (data->window);
+  else
+    unset_transparent_shape (data->window);
 }
 
 static void
@@ -172,9 +216,11 @@ timeline_finished_cb (GsdTimeline *timeline,
 {
   GsdLocatePointerData *data = (GsdLocatePointerData *) user_data;
 
-  /* hide window and unset shape */
+  /* set transparent shape and hide window */
+  if (!gtk_widget_is_composited (data->widget))
+    set_transparent_shape (data->window);
+
   gdk_window_hide (data->window);
-  gdk_window_shape_combine_mask (data->window, NULL, 0, 0);
 }
 
 static void
@@ -286,6 +332,12 @@ gsd_locate_pointer (GdkScreen *screen)
     }
 
   data->progress = 0.;
+
+  g_signal_connect (data->widget, "composited-changed",
+                    G_CALLBACK (composited_changed), data);
+
+  composited_changed (data->widget, data);
+
   gdk_window_show (data->window);
   move_locate_pointer_window (data, screen);
 
